#   Copyright ETH 2018 - 2024 Zürich, Scientific IT Services
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
#   
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import copy


def openbis_definitions(entity):
    """
    attrs_new: Attributes, that can appear when creating new entities
    attrs_up: Attributes that can be updated
    attrs: Attributes that are displayed when fetched
    multi: multivalue-elements which appear in an entity. E.g. parents or children in a Sample.
    identifier: to update entities, the identifier must be specified. Usually identityName + "Id"
    (Entity-Name in camel-case, starting with lowercase letter, with Id added)
    """
    entities = {
        "sessionInformation": {
            "attrs": "sessionToken userName homeGroupCode personalAccessTokenSession personalAccessTokenSessionName person creatorPerson".split(),
        },
        "space": {
            "attrs_new": "code description".split(),
            "attrs_up": "description freeze freezeForProjects freezeForSamples".split(),
            "attrs": "code permId description frozen frozenForProjects frozenForSamples registrator registrationDate modificationDate".split(),
            "multi": "".split(),
            "identifier": "spaceId",
            "create": {"@type": "as.dto.space.create.SpaceCreation"},
            "update": {"@type": "as.dto.space.update.SpaceUpdate"},
            "delete": {"@type": "as.dto.space.delete.SpaceDeletionOptions"},
            "fetch": {"@type": "as.dto.space.fetchoptions.SpaceFetchOptions"},
        },
        "project": {
            "attrs_new": "code description space attachments".split(),
            "attrs_up": "description space attachments freeze freezeForExperiments freezeForSamples".split(),
            "attrs": "code permId identifier space description leader registrator registrationDate modifier modificationDate attachments frozen frozenForExperiments frozenForSamples".split(),
            "multi": "".split(),
            "identifier": "projectId",
            "create": {"@type": "as.dto.project.create.ProjectCreation"},
            "update": {"@type": "as.dto.project.update.ProjectUpdate"},
        },
        "experiment": {
            "attrs_new": "code type project tags attachments metaData".split(),
            "attrs_up": "project tags attachments freeze freezeForDataSets freezeForSamples metaData immutableData".split(),
            "attrs": "code permId identifier type project tags registrator registrationDate modifier modificationDate attachments frozen frozenForDataSets frozenForSamples metaData immutableData".split(),
            "multi": "tags attachments".split(),
            "identifier": "experimentId",
            "create": {"@type": "as.dto.experiment.create.ExperimentCreation"},
            "update": {"@type": "as.dto.experiment.update.ExperimentUpdate"},
        },
        "externalDms": {
            "attrs_new": "code label address addressType creationId".split(),
            "attrs_up": "label address".split(),
            "attrs": "code permId label address addressType urlTemplate".split(),
            "identifier": "externalDmsId",
            "create": {"@type": "as.dto.externaldms.create.ExternalDmsCreation"},
            "update": {"@type": "as.dto.externaldms.update.ExternalDmsUpdate"},
        },
        "sample": {
            "attrs_new": "code type project parents children container components space experiment tags attachments metaData".split(),
            "attrs_up": "project parents children container components space experiment tags attachments freeze freezeForComponents freezeForChildren freezeForParents freezeForDataSets metaData immutableData".split(),
            "attrs": "code permId identifier type project parents children components space experiment tags registrator registrationDate modifier modificationDate attachments container frozen frozenForComponents frozenForChildren frozenForParents frozenForDataSets metaData immutableData".split(),
            "ids2type": {
                "parentIds": {"permId": {"@type": "as.dto.sample.id.SamplePermId"}},
                "childIds": {"permId": {"@type": "as.dto.sample.id.SamplePermId"}},
                "componentIds": {"permId": {"@type": "as.dto.sample.id.SamplePermId"}},
            },
            "identifier": "sampleId",
            "create": {"@type": "as.dto.sample.create.SampleCreation"},
            "update": {"@type": "as.dto.sample.update.SampleUpdate"},
            "delete": {"@type": "as.dto.sample.delete.SampleDeletionOptions"},
            "cre_type": "as.dto.sample.create.SampleCreation",
            "multi": "parents children components tags attachments".split(),
        },
        "sampleType": {
            "attrs_new": "code description autoGeneratedCode generatedCodePrefix subcodeUnique listable showContainer showParents showParentMetadata validationPlugin metaData".split(),
            "attrs_up": "description autoGeneratedCode generatedCodePrefix subcodeUnique listable showContainer showParents showParentMetadata, validationPlugin metaData".split(),
            "attrs": "permId code description autoGeneratedCode generatedCodePrefix subcodeUnique listable showContainer showParents showParentMetadata modificationDate validationPlugin metaData managedInternally".split(),
            "default_attribute_values": {
                "autoGeneratedCode": False,
                "generatedCodePrefix": "S",
                "subcodeUnique": False,
                "description": "",
                "listable": True,
                "showContainer": False,
                "showParents": True,
                "showParentMetadata": False,
            },
            "search": {"@type": "as.dto.sample.search.SampleTypeSearchCriteria"},
            "create": {"@type": "as.dto.sample.create.SampleTypeCreation"},
            "update": {"@type": "as.dto.sample.update.SampleTypeUpdate"},
            "delete": {"@type": "as.dto.sample.delete.SampleTypeDeletionOptions"},
            "identifier": "typeId",
        },
        "materialType": {
            "attrs_new": "code description validationPlugin".split(),
            "attrs_up": "description validationPlugin".split(),
            "attrs": "permId code description validationPlugin".split(),
            "search": {"@type": "as.dto.material.search.MaterialTypeSearchCriteria"},
            "fetch": {"@type": "as.dto.material.fetchoptions.MaterialTypeFetchOptions"},
            "create": {"@type": "as.dto.material.create.MaterialTypeCreation"},
            "update": {"@type": "as.dto.material.update.MaterialTypeUpdate"},
            "delete": {"@type": "as.dto.material.delete.MaterialTypeDeletionOptions"},
            "identifier": "typeId",
        },
        "dataSetType": {
            "attrs_new": "code description mainDataSetPattern mainDataSetPath disallowDeletion validationPlugin metaData".split(),
            "attrs_up": "description mainDataSetPattern mainDataSetPath disallowDeletion validationPlugin metaData".split(),
            "attrs": "permId code description mainDataSetPattern mainDataSetPath disallowDeletion modificationDate validationPlugin metaData managedInternally".split(),
            "search": {"@type": "as.dto.dataset.search.DataSetTypeSearchCriteria"},
            "fetch": {"@type": "as.dto.dataset.fetchoptions.DataSetTypeFetchOptions"},
            "create": {"@type": "as.dto.dataset.create.DataSetTypeCreation"},
            "update": {"@type": "as.dto.dataset.update.DataSetTypeUpdate"},
            "delete": {"@type": "as.dto.dataset.delete.DataSetTypeDeletionOptions"},
            "identifier": "typeId",
        },
        "personalAccessToken": {
            "attrs_new": "sessionName validFromDate validToDate accessDate".split(),
            "attrs_up": "".split(),
            "attrs": "permId sessionName validFromDate validToDate accessDate owner registrator registrationDate modifier modificationDate".split(),
            "search": {"@type": "as.dto.pat.search.PersonalAccessTokenSearchCriteria"},
            "delete": {"@type": "as.dto.pat.delete.PersonalAccessTokenDeletionOptions"},
            "identifier": "permId",
        },
        "experimentType": {
            "attrs_new": "code description validationPlugin metaData".split(),
            "attrs_up": "description modificationDate validationPlugin metaData".split(),
            "attrs": "permId code description modificationDate validationPlugin metaData managedInternally".split(),
            "search": {
                "@type": "as.dto.experiment.search.ExperimentTypeSearchCriteria"
            },
            "fetch": {
                "@type": "as.dto.experiment.fetchoptions.ExperimentTypeFetchOptions"
            },
            "create": {"@type": "as.dto.experiment.create.ExperimentTypeCreation"},
            "update": {"@type": "as.dto.experiment.update.ExperimentTypeUpdate"},
            "delete": {
                "@type": "as.dto.experiment.delete.ExperimentTypeDeletionOptions"
            },
            "identifier": "typeId",
        },
        "propertyType": {
            "attrs": "code label description managedInternally dataType vocabulary sampleType materialType schema transformation semanticAnnotations registrator registrationDate metaData multiValue".split(),
            "attrs_new": "code label description managedInternally dataType vocabulary sampleType materialType schema transformation metaData multiValue".split(),
            "attrs_up": "label description schema transformation metaData".split(),
            "search": {"@type": "as.dto.property.search.PropertyTypeSearchCriteria"},
            "create": {"@type": "as.dto.property.create.PropertyTypeCreation"},
            "update": {"@type": "as.dto.property.update.PropertyTypeUpdate"},
            "delete": {"@type": "as.dto.property.delete.PropertyTypeDeletionOptions"},
            "dataType": [
                "INTEGER",
                "VARCHAR",
                "MULTILINE_VARCHAR",
                "REAL",
                "TIMESTAMP",
                "DATE",
                "BOOLEAN",
                "CONTROLLEDVOCABULARY",
                "MATERIAL",
                "HYPERLINK",
                "XML",
                "SAMPLE",
                "ARRAY_INTEGER",
                "ARRAY_REAL",
                "ARRAY_STRING",
                "ARRAY_TIMESTAMP",
                "JSON"
            ],
            "identifier": "typeId",
        },
        "SemanticAnnotation": {
            "attrs_new": "permId entityType propertyType predicateOntologyId predicateOntologyVersion predicateAccessionId descriptorOntologyId descriptorOntologyVersion descriptorAccessionId".split(),
            "attrs_up": "entityType propertyType predicateOntologyId predicateOntologyVersion predicateAccessionId descriptorOntologyId descriptorOntologyVersion descriptorAccessionId ".split(),
            "attrs": "permId entityType propertyType predicateOntologyId predicateOntologyVersion predicateAccessionId descriptorOntologyId descriptorOntologyVersion descriptorAccessionId creationDate".split(),
            "ids2type": {
                "propertyTypeId": {"permId": "as.dto.property.id.PropertyTypePermId"},
                "entityTypeId": {"permId": "as.dto.entity.id.EntityTypePermId"},
            },
            "identifier": "permId",
            "cre_type": "as.dto.sample.create.SampleCreation",
            "multi": "parents children components tags attachments".split(),
        },
        "dataSet": {
            "attrs_new": "type code kind experiment sample parents children components containers tags metaData".split(),
            "attrs_up": "parents children experiment sample components containers tags freeze freezeForChildren freezeForParents freezeForComponents freezeForContainers metaData".split(),
            "attrs": "code permId type kind experiment sample parents children components containers tags accessDate dataProducer dataProductionDate registrator registrationDate modifier modificationDate dataStore measured postRegistered frozen frozenForChildren frozenForParents frozenForComponents frozenForContainers metaData".split(),
            "ids2type": {
                "parentIds": {"permId": {"@type": "as.dto.dataset.id.DataSetPermId"}},
                "childIds": {"permId": {"@type": "as.dto.dataset.id.DataSetPermId"}},
                "componentIds": {
                    "permId": {"@type": "as.dto.dataset.id.DataSetPermId"}
                },
                "containerIds": {
                    "permId": {"@type": "as.dto.dataset.id.DataSetPermId"}
                },
            },
            "multi": "parents children containers components".split(),
            "identifier": "dataSetId",
        },
        "material": {
            "attrs_new": "code description type creation tags".split(),
            "attrs_up": "description type creation tags".split(),
            "attrs": "code description type creation registrator registrationDate modifier modificationDate tags".split(),
            "multi": "".split(),
            "identifier": "materialId",
        },
        "tag": {
            "attrs_new": "code description".split(),
            "attrs_up": "description".split(),
            "attrs": "permId code description registrationDate owner".split(),
            "multi": "".split(),
            "identifier": "tagId",
        },
        "vocabulary": {
            "attrs_new": "code description managedInternally chosenFromList urlTemplate terms".split(),
            "attrs_up": "description managedInternally chosenFromList urlTemplate".split(),
            "attrs": "code description managedInternally chosenFromList urlTemplate registrator registrationDate modifier modificationDate".split(),
            "multi": "".split(),
            "identifier": "vocabularyId",
            "search": {"@type": "as.dto.vocabulary.search.VocabularySearchCriteria"},
            "create": {"@type": "as.dto.vocabulary.create.VocabularyCreation"},
            "update": {"@type": "as.dto.vocabulary.update.VocabularyUpdate"},
            "delete": {"@type": "as.dto.vocabulary.delete.VocabularyDeletionOptions"},
            "fetch": {"@type": "as.dto.vocabulary.fetchoptions.VocabularyFetchOptions"},
        },
        "vocabularyTerm": {
            "attrs_new": "code vocabularyCode label description official ordinal managedInternally".split(),
            "attrs_up": "label description official previousTermId".split(),
            "attrs": "code vocabularyCode label description official ordinal registrator registrationDate modifier modificationDate managedInternally".split(),
            "multi": "".split(),
            "identifier": "vocabularyTermId",
            "create": {"@type": "as.dto.vocabulary.create.VocabularyTermCreation"},
            "update": {"@type": "as.dto.vocabulary.update.VocabularyTermUpdate"},
            "delete": {
                "@type": "as.dto.vocabulary.delete.VocabularyTermDeletionOptions"
            },
            "fetch": {
                "@type": "as.dto.vocabulary.fetchoptions.VocabularyTermFetchOptions"
            },
        },
        "plugin": {
            "attrs_new": "name description pluginType script available entityKind".split(),
            "attrs_up": "description script available".split(),
            "attrs": "permId name description registrator registrationDate pluginKind entityKinds pluginType script available".split(),
            "multi": "".split(),
            "identifier": "pluginId",
            "pluginType": ["DYNAMIC_PROPERTY", "MANAGED_PROPERTY", "ENTITY_VALIDATION"],
            "entityKind": ["MATERIAL", "EXPERIMENT", "SAMPLE", "DATA_SET"],
        },
        "person": {
            "attrs_new": "userId space".split(),
            "attrs_up": "space".split(),
            "attrs": "permId userId firstName lastName email space registrationDate ".split(),
            "multi": "".split(),
            "identifier": "userId",
        },
        "authorizationGroup": {
            "attrs_new": "code description userIds".split(),
            "attrs_up": "code description userIds".split(),
            "attrs": "permId code description registrator registrationDate modificationDate users".split(),
            "multi": "users".split(),
            "identifier": "groupId",
        },
        "roleAssignment": {
            "attrs": "id user authorizationGroup role roleLevel space project registrator registrationDate".split(),
            "attrs_new": "role roleLevel user authorizationGroup role space project".split(),
            "attrs_up": "role roleLevel user authorizationGroup role space project".split(),
            "role": [
                "POWER_USER",
                "OBSERVER",
                "USER",
                "DISABLED",
                "ADMIN",
                "ETL_SERVER",
            ],
        },
        "attr2ids": {
            "space": "spaceId",
            "project": "projectId",
            "sample": "sampleId",
            "samples": "sampleIds",
            "sampleType": "sampleTypeId",
            "dataSet": "dataSetId",
            "dataSets": "dataSetIds",
            "experiment": "experimentId",
            "experiments": "experimentIds",
            "material": "materialId",
            "materials": "materialIds",
            "materialType": "materialTypeId",
            "container": "containerId",
            "containers": "containerIds",
            "component": "componentId",
            "components": "componentIds",
            "parents": "parentIds",
            "children": "childIds",
            "tags": "tagIds",
            "userId": "userId",
            "users": "userIds",
            "description": "description",
            "vocabulary": "vocabularyId",
            "validationPlugin": "validationPluginId",
            "metaData": "metaData",
        },
        "ids2type": {
            "spaceId": {"permId": {"@type": "as.dto.space.id.SpacePermId"}},
            "projectId": {"permId": {"@type": "as.dto.project.id.ProjectPermId"}},
            "experimentId": {
                "permId": {"@type": "as.dto.experiment.id.ExperimentPermId"}
            },
            "tagIds": {"code": {"@type": "as.dto.tag.id.TagCode"}},
        },
        "dataSetFile": {
            "search": {"@type": "dss.dto.datasetfile.search.DataSetFileSearchCriteria"}
        },
    }
    return entities[entity]


get_definition_for_entity = openbis_definitions  # Alias


fetch_option = {
    "personalAccessToken": {
        "@type": "as.dto.pat.fetchoptions.PersonalAccessTokenFetchOptions"
    },
    "space": {"@type": "as.dto.space.fetchoptions.SpaceFetchOptions"},
    "project": {
        "@type": "as.dto.project.fetchoptions.ProjectFetchOptions",
        "space": {"@type": "as.dto.space.fetchoptions.SpaceFetchOptions"},
    },
    "person": {"@type": "as.dto.person.fetchoptions.PersonFetchOptions"},
    "users": {"@type": "as.dto.person.fetchoptions.PersonFetchOptions"},
    "user": {"@type": "as.dto.person.fetchoptions.PersonFetchOptions"},
    "owner": {"@type": "as.dto.person.fetchoptions.PersonFetchOptions"},
    "registrator": {"@type": "as.dto.person.fetchoptions.PersonFetchOptions"},
    "modifier": {"@type": "as.dto.person.fetchoptions.PersonFetchOptions"},
    "leader": {"@type": "as.dto.person.fetchoptions.PersonFetchOptions"},
    "authorizationGroup": {
        "@type": "as.dto.authorizationgroup.fetchoptions.AuthorizationGroupFetchOptions",
         "roleAssignments": {
            "@type": "as.dto.roleassignment.fetchoptions.RoleAssignmentFetchOptions",
        }
    },
    "experiment": {
        "@type": "as.dto.experiment.fetchoptions.ExperimentFetchOptions",
        "type": {"@type": "as.dto.experiment.fetchoptions.ExperimentTypeFetchOptions"},
    },
    "sample": {
        "@type": "as.dto.sample.fetchoptions.SampleFetchOptions",
        "type": {"@type": "as.dto.sample.fetchoptions.SampleTypeFetchOptions"},
    },
    "samples": {"@type": "as.dto.sample.fetchoptions.SampleFetchOptions"},
    "sampleType": {
        "@type": "as.dto.sample.fetchoptions.SampleTypeFetchOptions",
        "propertyAssignments": {
            "@type": "as.dto.property.fetchoptions.PropertyAssignmentFetchOptions",
            "propertyType": {
                "@type": "as.dto.property.fetchoptions.PropertyTypeFetchOptions"
            },
            "plugin": {"@type": "as.dto.plugin.fetchoptions.PluginFetchOptions"},
        },
        "validationPlugin": {
            "@type": "as.dto.plugin.fetchoptions.PluginFetchOptions",
        },
    },
    "materialType": {
        "@type": "as.dto.material.fetchoptions.MaterialTypeFetchOptions",
        "propertyAssignments": {
            "@type": "as.dto.property.fetchoptions.PropertyAssignmentFetchOptions",
            "propertyType": {
                "@type": "as.dto.property.fetchoptions.PropertyTypeFetchOptions",
            },
            "plugin": {"@type": "as.dto.plugin.fetchoptions.PluginFetchOptions"},
        },
        "validationPlugin": {
            "@type": "as.dto.plugin.fetchoptions.PluginFetchOptions",
        },
    },
    "dataSetType": {
        "@type": "as.dto.dataset.fetchoptions.DataSetTypeFetchOptions",
        "propertyAssignments": {
            "@type": "as.dto.property.fetchoptions.PropertyAssignmentFetchOptions",
            "propertyType": {
                "@type": "as.dto.property.fetchoptions.PropertyTypeFetchOptions",
            },
            "plugin": {"@type": "as.dto.plugin.fetchoptions.PluginFetchOptions"},
        },
        "validationPlugin": {
            "@type": "as.dto.plugin.fetchoptions.PluginFetchOptions",
        },
    },
    "experimentType": {
        "@type": "as.dto.experiment.fetchoptions.ExperimentTypeFetchOptions",
        "propertyAssignments": {
            "@type": "as.dto.property.fetchoptions.PropertyAssignmentFetchOptions",
            "propertyType": {
                "@type": "as.dto.property.fetchoptions.PropertyTypeFetchOptions",
            },
            "plugin": {"@type": "as.dto.plugin.fetchoptions.PluginFetchOptions"},
        },
        "validationPlugin": {
            "@type": "as.dto.plugin.fetchoptions.PluginFetchOptions",
        },
    },
    "dataSet": {
        "@type": "as.dto.dataset.fetchoptions.DataSetFetchOptions",
        "type": {"@type": "as.dto.dataset.fetchoptions.DataSetTypeFetchOptions"},
        "parents": {"@type": "as.dto.dataset.fetchoptions.DataSetFetchOptions"},
        "children": {"@type": "as.dto.dataset.fetchoptions.DataSetFetchOptions"},
        "containers": {"@type": "as.dto.dataset.fetchoptions.DataSetFetchOptions"},
        "components": {"@type": "as.dto.dataset.fetchoptions.DataSetFetchOptions"},
    },
    "dataSets": {
        "@type": "as.dto.dataset.fetchoptions.DataSetFetchOptions",
        "properties": {"@type": "as.dto.property.fetchoptions.PropertyFetchOptions"},
        "type": {"@type": "as.dto.dataset.fetchoptions.DataSetTypeFetchOptions"},
    },
    "physicalData": {"@type": "as.dto.dataset.fetchoptions.PhysicalDataFetchOptions"},
    "linkedData": {
        "externalDms": {
            "@type": "as.dto.externaldms.fetchoptions.ExternalDmsFetchOptions"
        },
        "@type": "as.dto.dataset.fetchoptions.LinkedDataFetchOptions",
    },
    "roleAssignment": {
        "@type": "as.dto.roleassignment.fetchoptions.RoleAssignmentFetchOptions",
    },
    "roleAssignments": {
        "@type": "as.dto.roleassignment.fetchoptions.RoleAssignmentFetchOptions",
    },
    "properties": {"@type": "as.dto.property.fetchoptions.PropertyFetchOptions"},
    "propertyAssignments": {
        "@type": "as.dto.property.fetchoptions.PropertyAssignmentFetchOptions",
        "propertyType": {
            "@type": "as.dto.property.fetchoptions.PropertyTypeFetchOptions",
            "vocabulary": {
                "@type": "as.dto.vocabulary.fetchoptions.VocabularyFetchOptions",
            },
        },
    },
    "propertyType": {
        "@type": "as.dto.property.fetchoptions.PropertyTypeFetchOptions",
        "vocabulary": {
            "@type": "as.dto.vocabulary.fetchoptions.VocabularyFetchOptions"
        },
        "materialType": {
            "@type": "as.dto.material.fetchoptions.MaterialTypeFetchOptions"
        },
        "semanticAnnotations": {
            "@type": "as.dto.semanticannotation.fetchoptions.SemanticAnnotationFetchOptions"
        },
        "sampleType": {
            "@type": "as.dto.sample.fetchoptions.SampleTypeFetchOptions"
        },
        "registrator": {"@type": "as.dto.person.fetchoptions.PersonFetchOptions"},
    },
    "semanticAnnotations": {
        "@type": "as.dto.semanticannotation.fetchoptions.SemanticAnnotationFetchOptions"
    },
    "tags": {"@type": "as.dto.tag.fetchoptions.TagFetchOptions"},
    "tag": {"@type": "as.dto.tag.fetchoptions.TagFetchOptions"},
    "attachments": {"@type": "as.dto.attachment.fetchoptions.AttachmentFetchOptions"},
    "attachmentsWithContent": {
        "@type": "as.dto.attachment.fetchoptions.AttachmentFetchOptions",
        "content": {"@type": "as.dto.common.fetchoptions.EmptyFetchOptions"},
    },
    "script": {
        "@type": "as.dto.common.fetchoptions.EmptyFetchOptions",
    },
    "history": {"@type": "as.dto.history.fetchoptions.HistoryEntryFetchOptions"},
    "dataStore": {"@type": "as.dto.datastore.fetchoptions.DataStoreFetchOptions"},
    "plugin": {"@type": "as.dto.plugin.fetchoptions.PluginFetchOptions"},
    "vocabulary": {
        "@type": "as.dto.vocabulary.fetchoptions.VocabularyFetchOptions",
    },
    "vocabularyTerm": {
        "@type": "as.dto.vocabulary.fetchoptions.VocabularyTermFetchOptions"
    },
    "deletedObjects": {
        "@type": "as.dto.deletion.fetchoptions.DeletedObjectFetchOptions"
    },
    "deletion": {"@type": "as.dto.deletion.fetchoptions.DeletionFetchOptions"},
    "externalDms": {"@type": "as.dto.externaldms.fetchoptions.ExternalDmsFetchOptions"},
    "dataSetFile": {
        "@type": "dss.dto.datasetfile.fetchoptions.DataSetFileFetchOptions"
    },
}


def get_fetchoption_for_entity(entity):
    entity = entity[0].lower() + entity[1:]  # make first character lowercase
    try:
        return copy.deepcopy(fetch_option[entity])
    except KeyError as e:
        return {}


def get_type_for_entity(entity, action, parents_or_children=""):
    if action not in "create update delete search".split():
        raise ValueError(f"unknown action: {action}")

    definition = openbis_definitions(entity)
    if action in definition and not parents_or_children:
        return definition[action]
    else:
        # try to guess type, according to the naming scheme
        cap_entity = entity[:1].upper() + entity[1:]
        noun = {
            "create": "Creation",
            "update": "Update",
            "delete": "DeletionOptions",
            "search": "SearchCriteria",
        }

        if parents_or_children:
            return {
                "@type": f"as.dto.{entity.lower()}.{action}.{cap_entity}{parents_or_children}{noun[action]}"
            }
        else:
            return {
                "@type": f"as.dto.{entity.lower()}.{action}.{cap_entity}{noun[action]}"
            }


def get_fetchoptions(entity, including=None):
    if including is None:
        including = []
    including += ["@type"]
    entity = entity[0].lower() + entity[1:]  # make first character lowercase
    fo = {}
    for inc in including:
        try:
            item = fetch_option[entity][inc]
            fo[inc] = item
        except KeyError as err:
            pass
    return fo


def get_method_for_entity(entity: str, action: str) -> str:
    action = action.lower()

    if entity == "vocabulary":
        return f"{action}Vocabularies"

    cap_entity = entity[:1].upper() + entity[1:]

    return f"{action}{cap_entity}s"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   Copyright ETH 2018 - 2024 Zürich, Scientific IT Services
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
#   
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#



"""
data_set.py

Module with functions for operating on data sets.


Created by Chandrasekhar Ramakrishnan on 2017-04-05.
Copyright (c) 2017 Chandrasekhar Ramakrishnan. All rights reserved.
"""


def transfer_to_file_creation(content, file_creation, key, file_creation_key=None):
    if file_creation_key is None:
        file_creation_key = key
    if content.get(key) is not None:
        file_creation[file_creation_key] = content[key]


class GitDataSetCreation(object):
    def __init__(
        self,
        openbis,
        data_set_type,
        path,
        commit_id,
        repository_id,
        dms,
        sample=None,
        experiment=None,
        properties={},
        dss_code=None,
        parents=None,
        data_set_code=None,
        contents=[],
    ):
        """Initialize the command object with the necessary parameters.
        :param openbis: The openBIS API object.
        :param data_set_type: The type of the data set
        :param path: The path to the git repository
        :param commit_id: The git commit id
        :param repository_id: The git repository id - same for copies
        :param dms: An external data managment system object or external_dms_id
        :param sample: A sample object or sample id.
        :param experiment: An experiment or experiment id.
        :param properties: Properties for the data set.
        :param dss_code: Code for the DSS -- defaults to the first dss if none is supplied.
        :param parents: Parents for the data set.
        :param data_set_code: A data set code -- used if provided, otherwise generated on the server
        :param contents: A list of dicts that describe the contents:
            {'fileLength': [file length],
             'crc32': [crc32 checksum],
             'checksum': [checksum other than crc32],
             'checksumType': [checksum type if fiels checksum is used],
             'directory': [is path a directory?]
             'path': [the relative path string]}

        """
        self.openbis = openbis
        self.data_set_type = data_set_type
        self.path = path
        self.commit_id = commit_id
        self.repository_id = repository_id
        self.dms = dms
        self.sample = sample
        self.experiment = experiment
        self.properties = properties
        self.dss_code = dss_code
        self.parents = parents
        self.data_set_code = data_set_code
        self.contents = contents

    def new_git_data_set(self):
        """Create a link data set.
        :return: A DataSet object
        """

        data_set_creation = self.data_set_metadata_creation()
        file_metadata = self.data_set_file_metadata()
        if not file_metadata:
            return self.create_pure_metadata_data_set(data_set_creation)
        return self.create_mixed_data_set(data_set_creation, file_metadata)

    def create_pure_metadata_data_set(self, data_set_creation):
        # register the files in openBIS
        request = {
            "method": "createDataSets",
            "params": [self.openbis.token, [data_set_creation]],
        }

        # noinspection PyProtectedMember
        resp = self.openbis._post_request(self.openbis.as_v3, request)
        return self.openbis.get_dataset(resp[0]["permId"])

    def create_mixed_data_set(self, metadata_creation, file_metadata):
        data_set_creation = {
            "fileMetadata": file_metadata,
            "metadataCreation": metadata_creation,
            "@type": "dss.dto.dataset.create.FullDataSetCreation",
        }

        # register the files in openBIS
        request = {
            "method": "createDataSets",
            "params": [self.openbis.token, [data_set_creation]],
        }

        server_url = self.data_store_url(metadata_creation["dataStoreId"]["permId"])
        # noinspection PyProtectedMember
        resp = self.openbis._post_request_full_url(server_url, request)
        return self.openbis.get_dataset(resp[0]["permId"])

    def data_store_url(self, dss_code):
        data_stores = self.openbis.get_datastores()
        data_store = data_stores[data_stores["code"] == dss_code]
        return f'{data_store["downloadUrl"][0]}/datastore_server/rmi-data-store-server-v3.json'

    def data_set_metadata_creation(self):
        """Create the respresentation of the data set metadata."""
        dss_code = self.dss_code
        if dss_code is None:
            dss_code = self.openbis.get_datastores()["code"][0]

        dms_id = self.openbis.external_data_managment_system_to_dms_id(self.dms)
        parents = self.parents
        parentIds = []
        if parents is not None:
            if not isinstance(parents, list):
                parents = [parents]
            parentIds = [
                self.openbis.data_set_to_data_set_id(parent) for parent in parents
            ]
        data_set_creation = {
            "linkedData": {
                "@type": "as.dto.dataset.create.LinkedDataCreation",
                "contentCopies": [
                    {
                        "@type": "as.dto.dataset.create.ContentCopyCreation",
                        "path": self.path,
                        "gitCommitHash": self.commit_id,
                        "gitRepositoryId": self.repository_id,
                        "externalDmsId": dms_id,
                    }
                ],
            },
            "typeId": {
                "@type": "as.dto.entitytype.id.EntityTypePermId",
                "permId": self.data_set_type,
            },
            "dataStoreId": {
                "permId": dss_code,
                "@type": "as.dto.datastore.id.DataStorePermId",
            },
            "parentIds": parentIds,
            "measured": False,
            "properties": self.properties,
            "@type": "as.dto.dataset.create.DataSetCreation",
        }

        if self.sample is not None:
            sample_id = self.openbis.sample_to_sample_id(self.sample)
            data_set_creation["sampleId"] = sample_id
        elif self.experiment is not None:
            experiment_id = self.openbis.experiment_to_experiment_id(self.experiment)
            data_set_creation["experimentId"] = experiment_id

        if self.data_set_code is not None:
            data_set_creation["code"] = self.data_set_code
            data_set_creation["autoGeneratedCode"] = False
        else:
            data_set_creation["autoGeneratedCode"] = True

        return data_set_creation

    def data_set_file_metadata(self):
        """Create a representation of the file metadata"""
        return [self.as_file_metadata(c) for c in self.contents]

    def as_file_metadata(self, content):
        # The DSS objects do not use type
        # result = {"@type": "dss.dto.datasetfile.DataSetFileCreation"}
        result = {}
        transfer_to_file_creation(content, result, "fileLength")
        transfer_to_file_creation(content, result, "crc32", "checksumCRC32")
        transfer_to_file_creation(content, result, "checksum", "checksum")
        transfer_to_file_creation(content, result, "checksumType", "checksumType")
        transfer_to_file_creation(content, result, "directory")
        transfer_to_file_creation(content, result, "path")
        return result


class GitDataSetUpdate(object):
    def __init__(self, openbis, data_set_id):
        """Initialize the command object with the necessary parameters.
        :param openbis: The openBIS API object.
        :param data_set_id: Id of the data set to be updated
        """
        self.openbis = openbis
        self.data_set_id = data_set_id

    def new_content_copy(self, path, commit_id, repository_id, edms_id):
        """Create a data set update for adding a content copy.
        :return: A DataSetUpdate object
        """
        self.path = path
        self.commit_id = commit_id
        self.repository_id = repository_id
        self.edms_id = edms_id

        content_copy_actions = self.get_actions_add_content_copy()
        data_set_update = self.get_data_set_update(content_copy_actions)
        self.send_request(data_set_update)

    def delete_content_copy(self, content_copy):
        """Deletes the given content_copy from openBIS.
        :param content_copy: Content copy to be deleted.
        """
        content_copy_actions = self.get_actions_remove_content_copy(content_copy)
        data_set_update = self.get_data_set_update(content_copy_actions)
        self.send_request(data_set_update)

    def send_request(self, data_set_update):
        request = {
            "method": "updateDataSets",
            "params": [self.openbis.token, [data_set_update]],
        }
        self.openbis._post_request(self.openbis.as_v3, request)

    def get_data_set_update(self, content_copy_actions=[]):
        return {
            "@type": "as.dto.dataset.update.DataSetUpdate",
            "dataSetId": self.get_data_set_id(),
            "linkedData": self.get_linked_data(content_copy_actions),
        }

    def get_data_set_id(self):
        return {"@type": "as.dto.dataset.id.DataSetPermId", "permId": self.data_set_id}

    def get_linked_data(self, actions):
        return {
            "@type": "as.dto.common.update.FieldUpdateValue",
            "isModified": True,
            "value": {
                "@type": "as.dto.dataset.update.LinkedDataUpdate",
                "contentCopies": {
                    "@type": "as.dto.dataset.update.ContentCopyListUpdateValue",
                    "actions": actions,
                },
            },
        }

    def get_actions_add_content_copy(self):
        return [
            {
                "@type": "as.dto.common.update.ListUpdateActionAdd",
                "items": [self.get_content_copy_creation()],
            }
        ]

    def get_actions_remove_content_copy(self, content_copy):
        return [
            {
                "@type": "as.dto.common.update.ListUpdateActionRemove",
                "items": [content_copy["id"]],
            }
        ]

    def get_content_copy_creation(self):
        return {
            "@type": "as.dto.dataset.create.ContentCopyCreation",
            "externalDmsId": {
                "@type": "as.dto.externaldms.id.ExternalDmsPermId",
                "permId": self.edms_id,
            },
            "path": self.path,
            "gitCommitHash": self.commit_id,
            "gitRepositoryId": self.repository_id,
        }


class GitDataSetFileSearch(object):
    def __init__(self, openbis, data_set_id, dss_code=None):
        """Initialize the command object with the necessary parameters.
        :param openbis: The openBIS API object.
        :param data_set_id: Id of the data set to be updated
        :param dss_code: Code for the DSS -- defaults to the first dss if none is supplied.
        """
        self.openbis = openbis
        self.data_set_id = data_set_id
        self.dss_code = dss_code

    def search_files(self):
        request = {
            "method": "searchFiles",
            "params": [
                self.openbis.token,
                self.get_data_set_file_search_criteria(),
                self.get_data_set_file_fetch_options(),
            ],
        }
        server_url = self.data_store_url()
        return self.openbis._post_request_full_url(server_url, request)

    def get_data_set_file_search_criteria(self):
        return {
            "@type": "dss.dto.datasetfile.search.DataSetFileSearchCriteria",
            "operator": "AND",
            "criteria": [
                {
                    "@type": "as.dto.dataset.search.DataSetSearchCriteria",
                    "relation": "DATASET",
                    "operator": "OR",
                    "criteria": [
                        {
                            "fieldName": "code",
                            "fieldType": "ATTRIBUTE",
                            "fieldValue": {
                                "value": self.data_set_id,
                                "@type": "as.dto.common.search.StringEqualToValue",
                            },
                            "@type": "as.dto.common.search.CodeSearchCriteria",
                        }
                    ],
                }
            ],
        }

    def get_data_set_file_fetch_options(self):
        return {
            "@type": "dss.dto.datasetfile.fetchoptions.DataSetFileFetchOptions",
        }

    def data_store_url(self):
        data_stores = self.openbis.get_datastores()
        if self.dss_code is None:
            self.dss_code = self.openbis.get_datastores()["code"][0]
        data_store = data_stores[data_stores["code"] == self.dss_code]
        return f'{data_store["downloadUrl"][0]}/datastore_server/rmi-data-store-server-v3.json'



#   Copyright ETH 2018 - 2024 Zürich, Scientific IT Services
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
#   
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from tabulate import tabulate

from .utils import VERBOSE


class SemanticAnnotation:
    """Semantic annotation for sample types, property types or property type assignments."""

    def __init__(self, openbis_obj, isNew=True, **kwargs):
        self._openbis = openbis_obj
        self._isNew = isNew

        self.permId = kwargs.get("permId")
        self.entityType = kwargs.get("entityType")
        self.propertyType = kwargs.get("propertyType")
        self.predicateOntologyId = kwargs.get("predicateOntologyId")
        self.predicateOntologyVersion = kwargs.get("predicateOntologyVersion")
        self.predicateAccessionId = kwargs.get("predicateAccessionId")
        self.descriptorOntologyId = kwargs.get("descriptorOntologyId")
        self.descriptorOntologyVersion = kwargs.get("descriptorOntologyVersion")
        self.descriptorAccessionId = kwargs.get("descriptorAccessionId")
        self.creationDate = kwargs.get("creationDate")

    def __dir__(self):
        return [
            "permId",
            "entityType",
            "propertyType",
            "predicateOntologyId",
            "predicateOntologyVersion",
            "predicateAccessionId",
            "descriptorOntologyId",
            "descriptorOntologyVersion",
            "descriptorAccessionId",
            "creationDate",
            "save()",
            "delete()",
        ]

    def save(self):
        if self._isNew:
            self._create()
        else:
            self._update()

    def _create(self):

        creation = {
            "@type": "as.dto.semanticannotation.create.SemanticAnnotationCreation"
        }

        if self.entityType is not None and self.propertyType is not None:
            creation["propertyAssignmentId"] = {
                "@type": "as.dto.property.id.PropertyAssignmentPermId",
                "entityTypeId": {
                    "@type": "as.dto.entitytype.id.EntityTypePermId",
                    "permId": self.entityType,
                    "entityKind": "SAMPLE",
                },
                "propertyTypeId": {
                    "@type": "as.dto.property.id.PropertyTypePermId",
                    "permId": self.propertyType,
                },
            }
        elif self.entityType is not None:
            creation["entityTypeId"] = {
                "@type": "as.dto.entitytype.id.EntityTypePermId",
                "permId": self.entityType,
                "entityKind": "SAMPLE",
            }
        elif self.propertyType is not None:
            creation["propertyTypeId"] = {
                "@type": "as.dto.property.id.PropertyTypePermId",
                "permId": self.propertyType,
            }

        for attr in [
            "predicateOntologyId",
            "predicateOntologyVersion",
            "predicateAccessionId",
            "descriptorOntologyId",
            "descriptorOntologyVersion",
            "descriptorAccessionId",
        ]:
            creation[attr] = getattr(self, attr)

        request = {
            "method": "createSemanticAnnotations",
            "params": [self._openbis.token, [creation]],
        }

        response = self._openbis._post_request(self._openbis.as_v3, request)
        self._isNew = False
        self.permId = response[0]["permId"]

        if VERBOSE:
            print("Semantic annotation successfully created.")

    def _update(self):

        update = {
            "@type": "as.dto.semanticannotation.update.SemanticAnnotationUpdate",
            "semanticAnnotationId": {
                "@type": "as.dto.semanticannotation.id.SemanticAnnotationPermId",
                "permId": self.permId,
            },
        }

        for attr in [
            "predicateOntologyId",
            "predicateOntologyVersion",
            "predicateAccessionId",
            "descriptorOntologyId",
            "descriptorOntologyVersion",
            "descriptorAccessionId",
        ]:
            update[attr] = {
                "@type": "as.dto.common.update.FieldUpdateValue",
                "isModified": True,
                "value": getattr(self, attr),
            }

        request = {
            "method": "updateSemanticAnnotations",
            "params": [self._openbis.token, [update]],
        }

        self._openbis._post_request(self._openbis.as_v3, request)
        if VERBOSE:
            print("Semantic annotation successfully updated.")

    def delete(self, reason):
        self._openbis.delete_entity(
            entity="SemanticAnnotation", id=self.permId, reason=reason
        )
        if VERBOSE:
            print("Semantic annotation successfully deleted.")

    def _repr_html_(self):
        attrs = [
            "permId",
            "entityType",
            "propertyType",
            "predicateOntologyId",
            "predicateOntologyVersion",
            "predicateAccessionId",
            "descriptorOntologyId",
            "descriptorOntologyVersion",
            "descriptorAccessionId",
            "creationDate",
        ]

        html = """
            <table border="1" class="dataframe">
            <thead>
                <tr style="text-align: right;">
                <th>attribute</th>
                <th>value</th>
                </tr>
            </thead>
            <tbody>
        """

        for attr in attrs:
            html += f"<tr> <td>{attr}</td> <td>{getattr(self, attr, '')}</td> </tr>"

        html += """
            </tbody>
            </table>
        """
        return html

    def __repr__(self):
        headers = ["attribute", "value"]
        lines = []
        lines.append(["permId", self.permId])
        lines.append(["entityType", self.entityType])
        lines.append(["propertyType", self.propertyType])
        lines.append(["predicateOntologyId", self.predicateOntologyId])
        lines.append(["predicateOntologyVersion", self.predicateOntologyVersion])
        lines.append(["predicateAccessionId", self.predicateAccessionId])
        lines.append(["descriptorOntologyId", self.descriptorOntologyId])
        lines.append(["descriptorOntologyVersion", self.descriptorOntologyVersion])
        lines.append(["descriptorAccessionId", self.descriptorAccessionId])
        lines.append(["creationDate", self.creationDate])
        return tabulate(lines, headers=headers)



#   Copyright ETH 2018 - 2024 Zürich, Scientific IT Services
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
#   
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from .openbis_object import OpenBisObject


class Tag(OpenBisObject, entity="tag", single_item_method_name="get_tag"):
    """ """

    def __dir__(self):
        return [
            "get_samples()",
            "get_experiments()",
            "get_materials()",
            "get_owner()",
        ] + super().__dir__()

    def get_owner(self):
        return self.openbis.get_person(self.owner)

    def get_samples(self):
        return self.openbis.get_samples(tags=[self.code])
        # raise ValueError('not yet implemented')

    def get_experiments(self):
        return self.openbis.get_experiments(tags=[self.code])

    def get_materials(self):
        raise ValueError("not yet implemented")



#   Copyright ETH 2018 - 2023 Zürich, Scientific IT Services
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
#   
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
name = "pybis"
__author__ = "ID SIS • ETH Zürich"
__email__ = "openbis-support@id.ethz.ch"
__version__ = "6.7.0.0-rc0"

from . import pybis
from .pybis import DataSet
from .pybis import Openbis
from .pybis import ImagingControl
from .pybis import Spreadsheet
from .imaging import *



#   Copyright ETH 2018 - 2024 Zürich, Scientific IT Services
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
#   
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from pandas import DataFrame

from .openbis_object import OpenBisObject
from .things import Things
from .utils import VERBOSE, extract_permid, extract_nested_permid, format_timestamp


class Group(
    OpenBisObject, entity="authorizationGroup", single_item_method_name="get_group"
):
    """Managing openBIS authorization groups"""

    def __dir__(self):
        return [
            "code",
            "description",
            "users",
            "roleAssignments",
            "get_users()",
            "set_users()",
            "add_users()",
            "del_users()",
            "get_roles()",
            "assign_role()",
            "revoke_role(techId)",
        ]

    def get_persons(self):
        """Returns a Things object wich contains all Persons (Users)
        that belong to this group.
        """

        def create_data_frame(attrs, props, response):
            columns = [
                "permId",
                "userId",
                "firstName",
                "lastName",
                "email",
                "space",
                "registrationDate",
                "active",
            ]
            persons = DataFrame(response)
            if len(persons) == 0:
                persons = DataFrame(columns=columns)
            persons["permId"] = persons["permId"].map(extract_permid)
            persons["registrationDate"] = persons["registrationDate"].map(
                format_timestamp
            )
            persons["space"] = persons["space"].map(extract_nested_permid)

            return persons[columns]

        p = Things(
            self.openbis,
            entity="person",
            identifier_name="permId",
            response=self._users,
            df_initializer=create_data_frame,
        )
        return p

    get_users = get_persons  # Alias
    get_members = get_persons  # Alias

    def get_roles(self, **search_args):
        """Get all roles that are assigned to this group.
        Provide additional search arguments to refine your search.

        Usage::
            group.get_roles()
            group.get_roles(space='TEST_SPACE')
        """
        return self.openbis.get_role_assignments(group=self, **search_args)

    def assign_role(self, role, **kwargs):
        """Assign a role to this group. If no additional attribute is provided,
        roleLevel will default to INSTANCE. If a space attribute is provided,
        the roleLevel will be SPACE. If a project attribute is provided,
        roleLevel will be PROJECT.

        Usage::
            group.assign_role(role='ADMIN')
            group.assign_role(role='ADMIN', space='TEST_SPACE')

        """

        try:
            self.openbis.assign_role(role=role, group=self, **kwargs)
            if VERBOSE:
                print(f"Role {role} successfully assigned to group {self.code}")
        except ValueError as e:
            if "exists" in str(e):
                if VERBOSE:
                    print(f"Role {role} already assigned to group {self.code}")
            else:
                raise ValueError(str(e))

    def revoke_role(self, role, space=None, project=None, reason="no reason specified"):
        """Revoke a role from this group."""

        techId = None
        if isinstance(role, int):
            techId = role
        else:
            query = {"role": role}
            if space is None:
                query["space"] = ""
            else:
                query["space"] = space.upper()

            if project is None:
                query["project"] = ""
            else:
                query["project"] = project.upper()

            # build a query string for dataframe
            querystr = " & ".join(f'{key} == "{value}"' for key, value in query.items())
            roles = self.get_roles().df
            if len(roles) == 0:
                if VERBOSE:
                    print(
                        f"Role {role} has already been revoked from group {self.code}"
                    )
                return
            techId = roles.query(querystr)["techId"].values[0]

        # finally delete the role assignment
        ra = self.openbis.get_role_assignment(techId)
        ra.delete(reason)
        if VERBOSE:
            print(f"Role {role} successfully revoked from group {self.code}")
        return

    def save(self):
        if self.is_new:
            request = self._new_attrs()
            resp = self.openbis._post_request(self.openbis.as_v3, request)
            if VERBOSE:
                print("Group successfully created.")
            # re-fetch group from openBIS
            new_data = self.openbis.get_group(resp[0]["permId"], only_data=True)
            self._set_data(new_data)
            return self

        else:
            request = self._up_attrs()
            self.openbis._post_request(self.openbis.as_v3, request)
            if VERBOSE:
                print("Group successfully updated.")
            # re-fetch group from openBIS
            new_data = self.openbis.get_group(self.permId, only_data=True)
            self._set_data(new_data)



#   Copyright ETH 2018 - 2024 Zürich, Scientific IT Services
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
#   
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from .openbis_object import OpenBisObject
from .utils import (
    is_identifier,
    is_permid,
)


class Space(OpenBisObject, entity="space", single_item_method_name="get_space"):
    """managing openBIS spaces"""

    def __dir__(self):
        """all the available methods and attributes that should be displayed
        when using the autocompletion feature (TAB) in Jupyter
        """
        return [
            "get_projects()",
            "new_project()",
            "get_project()",
            "get_experiments()",
            "get_experiment()",
            "get_collections()",
            "get_collection()",
            "get_samples()",
            "get_sample()",
            "get_objects()",
            "get_object()",
            "delete()",
        ] + super().__dir__()

    def __str__(self):
        return self.data.get("code", None)

    def get_samples(self, **kwargs):
        return self.openbis.get_samples(space=self.code, **kwargs)

    get_objects = get_samples  # Alias

    def get_sample(self, sample_code, project_code=None):
        if is_identifier(sample_code) or is_permid(sample_code):
            return self.openbis.get_sample(sample_code)
        else:
            if project_code is None:
                return self.openbis.get_sample(f"/{self.code}/{sample_code}")
            else:
                return self.openbis.get_sample(
                    f"/{self.code}/{project_code}/{sample_code}"
                )

    get_object = get_sample  # Alias

    def get_project(self, project_code):
        if is_identifier(project_code):
            return self.openbis.get_project(project_code)
        else:
            projects = self.openbis.get_projects(code=project_code, space=self.code)
            if len(projects) == 0:
                raise ValueError(
                    f"No project named {project_code} in space {self.code}"
                )

            return projects[0]

    def get_projects(self, **kwargs):
        return self.openbis.get_projects(space=self.code, **kwargs)

    def new_project(self, code, description=None, **kwargs):
        return self.openbis.new_project(self.code, code, description, **kwargs)

    def get_experiments(self, **kwargs):
        return self.openbis.get_experiments(space=self.code, **kwargs)

    get_collections = get_experiments

    def get_experiment(self, experiment_code):
        if is_identifier(experiment_code):
            return self.openbis.get_experiment(experiment_code)
        else:
            experiments = self.openbis.get_experiments(
                code=experiment_code, space=self.code
            )
            if len(experiments) == 0:
                raise ValueError(
                    f"No project named {experiment_code} in space {self.code}"
                )

            return experiments[0]

    get_collection = get_experiment

    def new_sample(self, **kwargs):
        return self.openbis.new_sample(space=self, **kwargs)



#   Copyright ETH 2018 - 2024 Zürich, Scientific IT Services
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
#   
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from .openbis_object import OpenBisObject


class Experiment(
    OpenBisObject, entity="experiment", single_item_method_name="get_experiment"
):
    """Handling experiment (collection) entities in openBIS"""

    def _set_data(self, data):
        # assign the attribute data to self.a by calling it
        # (invoking the AttrHolder.__call__ function)
        self.a(data)
        self.__dict__["data"] = data

        # put the properties in the self.p namespace
        for key, value in data["properties"].items():
            property_type = self.p._property_names[key.lower()]
            data_type = property_type['dataType']
            if "multiValue" in property_type:
                if property_type['multiValue'] is True:
                    if type(value) is not list:
                        value = [value]
                    if data_type in (
                    "ARRAY_INTEGER", "ARRAY_REAL", "ARRAY_STRING", "ARRAY_TIMESTAMP"):
                        value = [self.formatter.to_array(data_type, x) for x in value]
                    else:
                        value = self.formatter.to_array(data_type, value)
                else:
                    if type(value) is list and data_type not in (
                    "ARRAY_INTEGER", "ARRAY_REAL", "ARRAY_STRING", "ARRAY_TIMESTAMP"):
                        raise ValueError(
                            f'Property type {property_type} is not a multi-value property!')
                    if data_type in (
                    "ARRAY_INTEGER", "ARRAY_REAL", "ARRAY_STRING", "ARRAY_TIMESTAMP"):
                        value = self.formatter.to_array(data_type, value)
            else:
                if data_type in ("ARRAY_INTEGER", "ARRAY_REAL", "ARRAY_STRING", "ARRAY_TIMESTAMP"):
                    value = self.formatter.to_array(data_type, value)
            if (data_type == 'XML' and 'metaData' in property_type and 'custom_widget' in property_type['metaData']
                    and property_type['metaData']['custom_widget'].upper() == 'SPREADSHEET'):
                    value = self.formatter.to_spreadsheet(value)
            self.p.__dict__[key.lower()] = value

    def __str__(self):
        return self.data["code"]

    def __dir__(self):
        # the list of possible methods/attributes displayed
        # when invoking TAB-completition
        return [
            "get_datasets()",
            "get_samples()",
            "set_tags()",
            "add_tags()",
            "del_tags()",
            "add_attachment()",
            "get_attachments()",
            "download_attachments()",
            "save()",
            "attrs",
            "props",
        ] + super().__dir__()

    @property
    def props(self):
        return self.__dict__["p"]

    @property
    def type(self):
        return self.__dict__["type"]

    @type.setter
    def type(self, type_name):
        experiment_type = self.openbis.get_experiment_type(type_name)
        self.p.__dict__["_type"] = experiment_type
        self.a.__dict__["_type"] = experiment_type

    def __getattr__(self, name):
        return getattr(self.__dict__["a"], name)

    def __setattr__(self, name, value):
        if name in ["set_properties", "add_tags()", "del_tags()", "set_tags()"]:
            raise ValueError("These are methods which should not be overwritten")
        elif name in ["p", "props"]:
            if isinstance(value, dict):
                for p in value:
                    setattr(self.__dict__["p"], p, value[p])
            else:
                raise ValueError("please provide a dictionary for setting properties")
        else:
            setattr(self.__dict__["a"], name, value)

    def _repr_html_(self):
        html = self.a._repr_html_()
        return html

    def set_properties(self, properties):
        for prop in properties.keys():
            setattr(self.p, prop, properties[prop])

    set_props = set_properties

    def get_datasets(self, **kwargs):
        if self.permId is None:
            return None
        return self.openbis.get_datasets(experiment=self.permId, **kwargs)

    def get_projects(self, **kwargs):
        if self.permId is None:
            return None
        return self.openbis.get_project(experiment=self.permId, **kwargs)

    def get_samples(self, **kwargs):
        if self.permId is None:
            return None
        return self.openbis.get_samples(experiment=self.permId, **kwargs)

    get_objects = get_samples  # Alias

    def add_samples(self, *samples):

        for sample in samples:
            if isinstance(sample, str):
                obj = self.openbis.get_sample(sample)
            else:
                # we assume we got a sample object
                obj = sample

            # a sample can only belong to exactly one experiment
            if obj.experiment is not None:
                raise ValueError(
                    f"sample {obj.code} already belongs to experiment {obj.experiment}"
                )
            else:
                if self.is_new:
                    raise ValueError(
                        "You need to save this experiment first before you can assign any samples to it"
                    )
                else:
                    # update the sample directly
                    obj.experiment = self.identifier
                    obj.save()
                    self.a.__dict__["_samples"].append(obj._identifier)

    add_objects = add_samples  # Alias

    def del_samples(self, samples):
        # TODO: implement this method using transactions

        raise ValueError("not implemented yet")
        if not isinstance(samples, list):
            samples = [samples]

        objects = []
        for sample in samples:
            if isinstance(sample, str):
                obj = self.openbis.get_sample(sample)
                objects.append(obj)
            else:
                # we assume we got an object
                objects.append(obj)
        self.samples = objects

    del_objects = del_samples  # Alias



#   Copyright ETH 2018 - 2024 Zürich, Scientific IT Services
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import base64
import os


class Attachment:
    def __init__(self, filename, title=None, description=None):
        if not os.path.exists(filename):
            raise ValueError(f"File not found: {filename}")
        self.fileName = filename
        self.title = title
        self.description = description

    def get_data_short(self):
        return {
            "fileName": self.fileName,
            "title": self.title,
            "description": self.description,
        }

    def get_data(self):
        with open(self.fileName, "rb") as att:
            content = att.read()
            contentb64 = base64.b64encode(content).decode()
        return {
            "fileName": self.fileName,
            "title": self.title,
            "description": self.description,
            "content": contentb64,
            "@type": "as.dto.attachment.create.AttachmentCreation",
        }



#   Copyright ETH 2018 - 2025 Zürich, Scientific IT Services
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
#   
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import math
import re
from datetime import datetime

# display messages when in a interactive context (IPython or Jupyter)
try:
    get_ipython()
except Exception:
    VERBOSE = False
else:
    VERBOSE = True


def parse_jackson(input_json):
    """openBIS uses a library called «jackson» to automatically generate the JSON RPC output.
    Objects that are found the first time are added an attribute «@id».
    Any further findings only carry this reference id.
    This function is used to dereference the output.
    """
    interesting = [
        "tags",
        "registrator",
        "modifier",
        "owner",
        "type",
        "parents",
        "children",
        "containers",  # 'container',
        "properties",
        "experiment",
        "sample",
        "project",
        "space",
        "propertyType",
        "entityType",
        "propertyType",
        "propertyAssignment",
        "externalDms",
        "roleAssignments",
        "user",
        "person",
        "creatorPerson",
        "users",
        "authorizationGroup",
        "vocabulary",
        "validationPlugin",
        "dataSetPermId",
        "dataStore",
        "sampleType"
    ]
    found = {}

    def build_cache(graph):
        if isinstance(graph, list):
            for item in graph:
                build_cache(item)
        elif isinstance(graph, dict) and len(graph) > 0:
            for key, value in graph.items():
                if key in interesting:
                    if isinstance(value, dict):
                        if "@id" in value:
                            found[value["@id"]] = value
                        build_cache(value)
                    elif isinstance(value, list):
                        for item in value:
                            if isinstance(item, dict):
                                if "@id" in item:
                                    found[item["@id"]] = item
                                build_cache(item)
                elif isinstance(value, dict):
                    build_cache(value)
                elif isinstance(value, list):
                    build_cache(value)

    def deref_graph(graph):
        if isinstance(graph, list):
            for i, list_item in enumerate(graph):
                if isinstance(list_item, int):
                    # try: # TODO: use "if list_item in found" with found.get()
                    #     graph[i] = found[list_item]
                    # except KeyError:
                    #     pass
                    if list_item in found:
                        graph[i] = found.get(list_item)
                else:
                    deref_graph(list_item)
        elif isinstance(graph, dict) and len(graph) > 0:
            for key, value in graph.items():
                if key in interesting:
                    if isinstance(value, dict):
                        deref_graph(value)
                    elif isinstance(value, int):
                        graph[key] = found.get(value)
                    elif isinstance(value, list):
                        for i, list_item in enumerate(value):
                            if isinstance(list_item, int):
                                if list_item in found:
                                    value[i] = found[list_item]
                                else:
                                    value[i] = list_item
                elif isinstance(value, dict):
                    deref_graph(value)
                elif isinstance(value, list):
                    deref_graph(value)

    build_cache(input_json)
    deref_graph(found)
    deref_graph(input_json)


def check_datatype(type_name, value, is_multi_value=False):
    if is_multi_value:
        if type_name == "INTEGER":
            return all([isinstance(x, int) and not math.isnan(x) for x in value])
        if type_name == "REAL":
            return all([isinstance(x, (int, float)) and not math.isnan(x) for x in value])
        if type_name == "BOOLEAN":
            return all([isinstance(x, bool) for x in value])
        if type_name == "VARCHAR":
            return all([isinstance(x, str) for x in value])
    else:
        if type_name == "INTEGER":
            return isinstance(value, int) and not math.isnan(value)
        if type_name == "REAL":
            return isinstance(value, (int, float)) and not math.isnan(value)
        if type_name == "BOOLEAN":
            return isinstance(value, bool)
        if type_name == "VARCHAR":
            return isinstance(value, str)
        if type_name is not None and type_name.startswith("ARRAY"):
            return isinstance(value, list)
    return True


def split_identifier(ident):
    bla = []
    bla = ident.upper().split("/")
    results = {}
    try:
        if bla[0] == "":
            bla.pop(0)
        if bla[-1] == "":
            bla.pop(-1)
        results["space"] = bla.pop(0)
        results["code"] = bla.pop(-1)
        results["experiment"] = bla.pop(0)
    except Exception:
        pass

    return results


def format_timestamp(ts):
    if ts is None:
        return ""
    if ts != ts:  # test for NaN
        return ""
    return datetime.fromtimestamp(round(ts / 1000)).strftime("%Y-%m-%d %H:%M:%S")


def is_identifier(ident):
    # assume we got a sample identifier e.g. /TEST/TEST-SAMPLE
    match = re.search("/", ident)
    if match:
        return True
    else:
        return False


def is_permid(ident):
    match = re.match("^\d+\-\d+$", ident)
    if match:
        return True
    else:
        return False


def nvl(val, string=""):
    if val is None:
        return string
    return val


def extract_permid(permid):
    if not isinstance(permid, dict):
        return str(permid)
    return permid["permId"]


def extract_data_type(obj):
    if not isinstance(obj, dict):
        return "" if obj is None else str(obj)
    return "" if obj["dataType"] is None else obj["dataType"]


def extract(obj, property_name):
    if not isinstance(obj, dict):
        return "" if obj is None else str(obj)
    return "" if obj[property_name] is None else obj[property_name]


def extract_code(obj):
    if not isinstance(obj, dict):
        return "" if obj is None else str(obj)
    return "" if obj["code"] is None else obj["code"]


def extract_downloadUrl(obj):
    if not isinstance(obj, dict):
        return "" if obj is None else str(obj)
    return "" if obj["downloadUrl"] is None else obj["downloadUrl"]


def extract_name(obj):
    if not isinstance(obj, dict):
        return "" if obj is None else str(obj)
    return "" if obj["name"] is None else obj["name"]


def extract_deletion(obj):
    del_objs = []
    for deleted_object in obj["deletedObjects"]:
        del_objs.append(
            {
                "reason": obj["reason"],
                "permId": deleted_object["id"]["permId"],
                "type": deleted_object["id"]["@type"],
                "deletionId": obj["id"]["id"]
            }
        )
    return del_objs


def extract_attr(attr):
    def attr_func(obj):
        if isinstance(obj, dict):
            return obj.get(attr, "")
        else:
            return str(obj)

    return attr_func


def extract_identifier(ident):
    if not isinstance(ident, dict):
        return "" if ident is None else str(ident)
    return "" if ident["identifier"] is None else ident["identifier"]


def extract_identifiers(items):
    if not items:
        return []
    try:
        return [
            data["identifier"]["identifier"]
            if "identifier" in data
            else data["permId"]["permId"]
            for data in items
        ]
    except TypeError:
        return []


def extract_nested_identifier(ident):
    if not isinstance(ident, dict):
        return "" if ident is None else str(ident)
    return "" if ident["identifier"]["identifier"] is None else ident["identifier"]["identifier"]


def extract_nested_permid(permid):
    if not isinstance(permid, dict):
        return "" if permid is None else str(permid)
    return "" if permid["permId"]["permId"] is None else permid["permId"]["permId"]


def extract_nested_permids(items):
    if not isinstance(items, list):
        return []

    return list(item["permId"]["permId"] for item in items)


def extract_property_assignments(pas):
    pa_strings = []
    for pa in pas:
        if not isinstance(pa["propertyType"], dict):
            pa_strings.append(pa["propertyType"])
        else:
            pa_strings.append(pa["propertyType"]["label"])
    return pa_strings


def extract_role_assignments(ras):
    ra_strings = []
    for ra in ras:
        ra_strings.append(
            {
                "role": ra["role"],
                "roleLevel": ra["roleLevel"],
                "space": ra["space"]["code"] if ra["space"] else None,
            }
        )
    return ra_strings


def extract_person(person):
    if not isinstance(person, dict):
        if person is None:
            return ""
        else:
            return str(person)
    return person["userId"]


def extract_person_details(person):
    if not isinstance(person, dict):
        return str(person)
    return "{firstName} {lastName} <{email}>".format(**person)


def extract_id(id):
    if not isinstance(id, dict):
        return str(id)
    else:
        return id["techId"]


def extract_userId(user):
    if isinstance(user, list):
        return ", ".join([u["userId"] for u in user])
    elif isinstance(user, dict):
        return "" if user["userId"] is None else user["userId"]
    else:
        return "" if user is None else str(user)


def is_number(value):
    """detects whether a given value
    is either an integer or a floating point number
    1, 2, 3,  etc.
    1.0, 2.1, etc.
    .5, .6    etc.
    """
    number_regex = re.compile(r"^(?=.)([+-]?([0-9]*)(\.([0-9]+))?)$")
    return number_regex.search(value)


def extract_username_from_token(token: str) -> str:
    if token.startswith("$pat"):
        match = re.search(r"^\$pat-(?P<username>.*?)-.*", token)
    else:
        match = re.search(r"(?P<username>.*?)-.*", token)
    if match:
        return match.groupdict().get("username")



#   Copyright ETH 2018 - 2024 Zürich, Scientific IT Services
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
#   
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from tabulate import tabulate

from .utils import (
    check_datatype
)


class PropertyHolder:
    def __init__(self, openbis_obj, type=None):
        self.__dict__["_openbis"] = openbis_obj
        self.__dict__["_property_names"] = {}
        if type is None:
            return

        self.__dict__["_type"] = type
        if (
                "propertyAssignments" in type.data
                and type.data["propertyAssignments"] is not None
        ):
            for prop in type.data["propertyAssignments"]:
                property_name = prop["propertyType"]["code"].lower()
                self._property_names[property_name] = prop["propertyType"]
                self._property_names[property_name]["mandatory"] = prop["mandatory"]
                self._property_names[property_name]["showInEditView"] = prop[
                    "showInEditView"
                ]
                if prop["propertyType"]["dataType"] == "CONTROLLEDVOCABULARY":
                    pt = self._openbis.get_property_type(prop["propertyType"]["code"])
                    # get the vocabulary of a property type.
                    # In some cases, the «code» of an assigned property is not identical to the «vocabulary» attribute
                    voc = self._openbis.get_vocabulary(pt.vocabulary)
                    terms = voc.get_terms()
                    self._property_names[property_name]["terms"] = terms

    def _all_props(self):
        props = {}
        if not getattr(self, "_type"):
            return props
        for code in self._type.codes():
            props[code] = getattr(self, code)
        return props

    def all(self):
        """Returns the properties as an array"""
        props = {}
        for code in self._type.codes():
            props[code] = getattr(self, code)
        return props

    def all_nonempty(self):
        props = {}
        for code in self._type.codes():
            value = getattr(self, code)
            if value is not None:
                props[code] = value
        return props

    def __call__(self, *args):
        if len(args) == 0:
            return self.all()
        elif len(args) == 1:
            return getattr(self, args[0])
        elif len(args) == 2:
            return setattr(self, args[0], args[1])
        else:
            raise ValueError("called properties with more than 2 arguments")

    def get(self, *args):
        if len(args) == 0:
            return self.all()
        elif len(args) == 1 and not isinstance(args[0], list):
            return getattr(self, args[0])
        else:
            if isinstance(args[0], list):
                args = args[0]
            return {arg: getattr(self, arg, None) for arg in args}

    def set(self, *args):
        if len(args) == 2:
            setattr(self, args[0], args[1])
        elif len(args) == 1 and isinstance(args[0], dict):
            for key in args[0]:
                setattr(self, key, args[0][key])

    def __getitem__(self, key):
        """For properties that contain either a dot or a dash or any other non-valid method character,
        a user can use a key-lookup instead, e.g. sample.props['my-weird.property-name']
        """

        return getattr(self, key)

    def __getattr__(self, name):
        """attribute syntax can be found out by
        adding an underscore at the end of the property name
        """
        if name == "_ipython_canary_method_should_not_exist_":
            # make Jupyter use the _repr_html_ method
            return
        if name.endswith("_"):
            name = name.rstrip("_")
            if name in self._property_names:
                property_type = self._property_names[name]
                if property_type["dataType"] == "CONTROLLEDVOCABULARY":
                    return property_type["terms"]
                    # return self._get_terms(property_type['code'])
                else:
                    syntax = {property_type["label"]: property_type["dataType"]}
                    if property_type["dataType"] == "TIMESTAMP":
                        syntax["syntax"] = "YYYY-MM-DD HH:MIN:SS"
                    return syntax
            else:
                return

    def __setattr__(self, name, value):
        """This special method allows a PropertyHolder object
        to check the attributes that are assigned to that object
        """
        if name not in self._property_names:
            raise KeyError(
                f"No such property: «{name}». Allowed properties are: {', '.join(self._property_names.keys())}"
            )
        property_type = self._property_names[name]
        data_type = property_type["dataType"]
        if "multiValue" in property_type and property_type["multiValue"] is not True and type(
                value) == list and data_type.startswith('ARRAY_') is False:
            raise ValueError(
                f'Property type {property_type["code"]} is not a multi-value property!')
        if value == '':
            value = None
        if value is not None:
            if data_type == "CONTROLLEDVOCABULARY":
                terms = property_type["terms"]
                if "multiValue" in property_type and property_type["multiValue"] is True:
                    if type(value) != list:
                        value = [value]
                    for single_value in value:
                        if str(single_value).upper() not in terms.df["code"].values:
                            raise ValueError(
                                f"Value for attribute «{name}» must be one of these terms: {', '.join(terms.df['code'].values)}"
                            )
                else:
                    value = str(value).upper()
                    if value not in terms.df["code"].values:
                        raise ValueError(
                            f"Value for attribute «{name}» must be one of these terms: {', '.join(terms.df['code'].values)} VALUE:{value}"
                        )
            elif data_type == "SAMPLE":
                if "multiValue" in property_type and property_type["multiValue"] is True:
                    if type(value) != list:
                        value = [value]
            elif data_type in ("INTEGER", "BOOLEAN", "VARCHAR", "REAL", "ARRAY_INTEGER", "ARRAY_REAL", "ARRAY_STRING"):
                is_multi_value = property_type["multiValue"] is True if "multiValue" in property_type else False
                if not check_datatype(data_type, value, is_multi_value):
                    if is_multi_value:
                        raise ValueError(f"Multi-value property '{property_type['code']}' must be of type {data_type} - Provided value:{value}")
                    else:
                        raise ValueError(f"Property '{property_type['code']}' must be of type {data_type} - Provided value:{value}")
        self.__dict__[name] = value

    def __setitem__(self, key, value):
        """For properties that contain either a dot or a dash or any other non-valid method character,
        a user can use a key instead, e.g. sample.props['my-weird.property-name']
        """
        return setattr(self, key, value)

    def __dir__(self):
        return self._property_names

    def _repr_html_(self):
        def nvl(val, string=""):
            if val is None:
                return string
            elif val == "true":
                return True
            elif val == "false":
                return False
            return val

        html = """
            <table border="1" class="dataframe">
            <thead>
                <tr style="text-align: right;">
                <th>property</th>
                <th>value</th>
                <th>description</th>
                <th>type</th>
                <th>mandatory</th>
                </tr>
            </thead>
            <tbody>
        """

        for prop_name, prop in self._property_names.items():
            html += "<tr>"
            html += "".join(
                f"<td>{item}</td>"
                for item in [
                    prop_name,
                    nvl(getattr(self, prop_name, ""), ""),
                    prop.get("description"),
                    prop.get("dataType"),
                    prop.get("mandatory"),
                ]
            )
            html += "</tr>"

        html += """
            </tbody>
            </table>
        """
        return html

    def __repr__(self):
        def nvl(val, string=""):
            if val is None:
                return string
            elif val == "true":
                return True
            elif val == "false":
                return False
            return str(val)

        headers = ["property", "value", "mandatory"]

        lines = []
        for prop_name in self._property_names:
            lines.append([prop_name, nvl(getattr(self, prop_name, ""))])
        return tabulate(lines, headers=headers)



#   Copyright ETH 2023 - 2024 Zürich, Scientific IT Services
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from datetime import datetime

import pandas as pd
import base64
import json

from .spreadsheet import Spreadsheet


def is_of_openbis_supported_date_format(value):
    is_supported = False
    for date_format in PropertyReformatter.SUPPORTED_DATETIME_FORMATS:
        try:
            datetime.strptime(value, date_format)
            is_supported = True
        except ValueError:
            pass
    return is_supported


class PropertyReformatter:
    """Helper class for reformatting of properties, is needed"""
    LONG_DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"

    SUPPORTED_DATETIME_FORMATS = ["%Y-%m-%d", "%y-%m-%d",  # ShortDateFormat
                                  "%Y-%m-%d %H:%M", "%y-%m-%d %H:%M",  # NormalDateFormat
                                  "%Y-%m-%d %H:%M:%S", "%y-%m-%d %H:%M:%S"]  # LongDateFormat

    def __init__(self, openbis_obj):
        self.openbis = openbis_obj

    def format(self, properties):
        if properties is None:
            raise ValueError('properties can not be None!')

        for key, value in properties.items():
            if value is None or value == '':
                properties[key] = None
                continue
            property_type = self.openbis.get_property_type(key)
            if property_type.dataType == 'TIMESTAMP':
                properties[key] = self._format_timestamp(value)
            elif property_type.dataType == 'ARRAY_TIMESTAMP':
                if property_type.multiValue:
                    properties[key] = ["[" + ",".join(map(str, [self._format_timestamp(x) for x in arr])) + "]" for arr in value]
                else:
                    properties[key] = [self._format_timestamp(x) for x in value]
            elif property_type.dataType.startswith('ARRAY'):
                if property_type.multiValue:
                    properties[key] = ["[" + ",".join(map(str, x)) + "]" for x in value]
            elif (property_type.dataType == 'XML' and 'custom_widget' in property_type.metaData
                    and property_type.metaData['custom_widget'].upper() == 'SPREADSHEET'):
                    if isinstance(value, Spreadsheet):
                        json_str = value.to_json().encode('utf-8')
                        b64 = base64.b64encode(json_str).decode('utf-8')
                        result = f'<DATA>{b64}</DATA>'
                        properties[key] = result
        return properties

    def _format_timestamp(self, value):
        if value is None:
            return value
        if is_of_openbis_supported_date_format(value):
            return value
        timestamp = pd.to_datetime(value)
        result = timestamp.strftime(PropertyReformatter.LONG_DATETIME_FORMAT)
        print(
            f'WARNING: "{value}" is not of any OpenBis supported datetime formats. Reformatting to "{result}"')
        return result

    def to_array(self, data_type, prop_value):
        if prop_value is None or prop_value == "":
            return []
        result = []
        if data_type in ("ARRAY_INTEGER", "INTEGER"):
            result = [int(x.strip()) for x in prop_value]
        elif data_type in ("ARRAY_REAL", "REAL"):
            result = [float(x.strip()) for x in prop_value]
        elif data_type == "BOOLEAN":
            result = [x.strip().lower() == "true" for x in prop_value]
        elif data_type in ("ARRAY_TIMESTAMP", "TIMESTAMP", "DATE"):
            result = [x.strip() for x in prop_value]
        else:
            result = prop_value
        return result

    def to_spreadsheet(self, rawValue):
        try:
            b64 = rawValue[len("<DATA>"):-len("</DATA>")]
            jsonb = base64.b64decode(b64)
            try:
                json_str = jsonb.decode('utf-8')
            except UnicodeDecodeError as decode_error:
                json_str = jsonb.decode('latin1')
            result = json.loads(json_str)
            return Spreadsheet.from_dict(result)
        except ValueError as e:
            print(f"Could not decode spreadsheet property: {e}")
            return rawValue





#   Copyright ETH 2018 - 2024 Zürich, Scientific IT Services
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
#   
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from .attribute import AttrHolder
from .definitions import openbis_definitions, get_type_for_entity, get_method_for_entity
from .openbis_object import OpenBisObject
from .utils import VERBOSE


class Vocabulary(
    OpenBisObject, entity="vocabulary", single_item_method_name="get_vocabulary"
):
    def __init__old_(self, openbis_obj, data=None, terms=None, **kwargs):
        self.__dict__["openbis"] = openbis_obj
        self.__dict__["a"] = AttrHolder(openbis_obj, "vocabulary")

        if data is not None:
            self._set_data(data)
            self.__dict__["terms"] = data["terms"]

        if terms is None:
            self.__dict__["terms"] = []
        else:
            self.__dict__["terms"] = terms

        if self.is_new:
            allowed_attrs = openbis_definitions(self.entity)["attrs_new"]
            for key in kwargs:
                if key not in allowed_attrs:
                    raise ValueError(
                        f"{key} is an unknown Vocabulary attribute. Allowed attributes are: {', '.join(allowed_attrs)}"
                    )

        if kwargs is not None:
            for key in kwargs:
                setattr(self, key, kwargs[key])

    def __dir__(self):
        return [
            "get_terms()",
            "add_term(code, label, description)",
            "save()",
        ]

    def get_terms(self):
        """Returns the VocabularyTerms of the given Vocabulary."""
        return self.openbis.get_terms(vocabulary=self.code)

    def add_term(self, code, label=None, description=None):
        """Adds a term to this Vocabulary.
        If Vocabulary is already persistent, it is added by adding a new VocabularyTerm object.
        If Vocabulary is new, the term is added to the list of terms
        """
        self.__dict__["terms"].append(
            {"code": code, "label": label, "description": description}
        )

    def delete(self, reason):
        """Delete this vocabulary"""
        if not self.data:
            return

        delete_type = get_type_for_entity("vocabulary", "delete")
        method = get_method_for_entity("vocabulary", "delete")

        request = {
            "method": method,
            "params": [
                self.openbis.token,
                [
                    {
                        "permId": self.code,
                        "@type": "as.dto.vocabulary.id.VocabularyPermId",
                    }
                ],
                {"reason": reason, **delete_type},
            ],
        }
        resp = self.openbis._post_request(self.openbis.as_v3, request)
        if VERBOSE:
            print(f"{self.entity} {self.code} successfully deleted.")

    def save(self):
        terms = self._terms or []
        for term in terms:
            term["@type"] = "as.dto.vocabulary.create.VocabularyTermCreation"

        if self.is_new:
            request = self._new_attrs("createVocabularies")
            if terms:
                request["params"][1][0]["terms"] = terms
            resp = self.openbis._post_request(self.openbis.as_v3, request)

            if VERBOSE:
                print("Vocabulary successfully created.")
            data = self.openbis.get_vocabulary(resp[0]["permId"], only_data=True)
            self._set_data(data)
            return self

        else:
            request = self._up_attrs("updateVocabularies")
            request["params"][1][0]["vocabularyId"] = {
                "@type": "as.dto.vocabulary.id.VocabularyPermId",
                "permId": self.code,
            }
            request["params"][1][0]["chosenFromList"] = {
                "value": self.chosenFromList,
                "isModified": True,
                "@type": "as.dto.common.update.FieldUpdateValue",
            }
            if terms:
                request["params"][1][0]["terms"] = terms
            self.openbis._post_request(self.openbis.as_v3, request)
            if VERBOSE:
                print("Vocabulary successfully updated.")
            data = self.openbis.get_vocabulary(self.code, only_data=True)
            self._set_data(data)


class VocabularyTerm(OpenBisObject):
    def __init__(self, openbis_obj, data=None, **kwargs):
        self.__dict__["openbis"] = openbis_obj
        self.__dict__["a"] = AttrHolder(openbis_obj, "vocabularyTerm")

        if data is not None:
            self._set_data(data)

        if kwargs is not None:
            for key in kwargs:
                setattr(self, key, kwargs[key])

    @property
    def vocabularyCode(self):
        if self.is_new:
            return self.__dict__["a"].vocabularyCode
        else:
            return self.data["permId"]["vocabularyCode"]

    def __dir__(self):
        return [
            "code",
            "vocabularyCode",
            "label",
            "description",
            "official",
            "ordinal",
            "registrator",
            "registrationDate",
            "modifier",
            "modificationDate",
            "move_to_top()",
            "move_after_term()",
        ]

    def move_to_top(self):
        """Moves the term on the top of the vocabularyTerm list,
        i.e. the ordinal will change
        """
        self.previousTermId = ""

    def move_after_term(self, term):
        """Moves the term just after the term given. This will result in an ordinal change."""
        self.previousTermId = term

    def _up_attrs(self):
        """AttributeTerms behave quite differently to all other openBIS entities,
        that's why we need to override this method
        """
        attrs = {}
        for attr in "label description official".split():
            attrs[attr] = {
                "value": getattr(self, attr),
                "isModified": True,
                "@type": "as.dto.common.update.FieldUpdateValue",
            }

        if not getattr(self, "previousTermId") == None:
            # internal attribute previousTermId has been touched:
            # set value to either None (i.e. move it to the top) or
            # set value to PREVIOUS_TERM like this:
            value = self.previousTermId
            if value == "":
                value = None
            else:
                permId = {
                    "@type": "as.dto.vocabulary.id.VocabularyTermPermId",
                    "vocabularyCode": self.vocabularyCode,
                    "code": value,
                }
                value = permId

            attrs["previousTermId"] = {
                "isModified": True,
                "@type": "as.dto.common.update.FieldUpdateValue",
                "value": value,
            }

        attrs["vocabularyTermId"] = self.vocabularyTermId()
        attrs["@type"] = "as.dto.vocabulary.update.VocabularyTermUpdate"
        request = {
            "method": "updateVocabularyTerms",
            "params": [self.openbis.token, [attrs]],
        }
        return request

    def _new_attrs(self):
        attrs = {
            "@type": "as.dto.vocabulary.create.VocabularyTermCreation",
            "vocabularyId": self.vocabularyTermId(),
        }
        for attr in "code label description".split():
            attrs[attr] = getattr(self, attr)

        request = {
            "method": "createVocabularyTerms",
            "params": [self.openbis.token, [attrs]],
        }
        return request

    def vocabularyTermId(self):
        """needed for updating a term."""
        if self.is_new:
            return {
                "permId": getattr(self, "vocabularyCode"),
                "@type": "as.dto.vocabulary.id.VocabularyPermId",
            }
        else:
            permId = self.data["permId"]
            permId.pop("@id", None)
            return permId

    def save(self):
        if self.is_new:
            request = self._new_attrs()
            resp = self.openbis._post_request(self.openbis.as_v3, request)

            if VERBOSE:
                print("Vocabulary Term successfully created.")
            data = self.openbis.get_term(
                code=resp[0]["code"],
                vocabularyCode=resp[0]["vocabularyCode"],
                only_data=True,
            )
            self._set_data(data)
            return self

        else:
            request = self._up_attrs()
            self.openbis._post_request(self.openbis.as_v3, request)
            if VERBOSE:
                print("Vocabulary Term successfully updated.")
            data = self.openbis.get_term(
                code=self.code, vocabularyCode=self.vocabularyCode, only_data=True
            )
            self._set_data(data)

    def delete(self, reason="no particular reason"):
        self.openbis.delete_openbis_entity(
            entity="vocabularyTerm", objectId=self.data["permId"], reason=reason
        )
        if VERBOSE:
            print("VocabularyTerm successfully deleted.")



#   Copyright ETH 2018 - 2024 Zürich, Scientific IT Services
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
#   
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from .attribute import AttrHolder
from .openbis_object import OpenBisObject
from .property import PropertyHolder
from .property_reformatter import PropertyReformatter
from .utils import VERBOSE


class Material(OpenBisObject):
    """Managing openBIS materials"""

    def __init__(self, openbis_obj, type, data=None, props=None, **kwargs):
        self.__dict__["entity"] = "material"
        self.__dict__["openbis"] = openbis_obj
        self.__dict__["type"] = type
        ph = PropertyHolder(openbis_obj, type)
        self.__dict__["p"] = ph
        self.__dict__["props"] = ph
        self.__dict__["a"] = AttrHolder(openbis_obj, "material", type)
        self.__dict__["formatter"] = PropertyReformatter(openbis_obj)

        if data is not None:
            self._set_data(data)

        if props is not None:
            for key in props:
                setattr(self.p, key, props[key])

        if kwargs is not None:
            for key in kwargs:
                setattr(self, key, kwargs[key])

    def __dir__(self):
        return ["code", "description", "set_tags()", "add_tags()", "del_tags()"]

    def save(self):
        for prop_name, prop in self.props._property_names.items():
            if prop["mandatory"]:
                if (
                    getattr(self.props, prop_name) is None
                    or getattr(self.props, prop_name) == ""
                ):
                    raise ValueError(
                        f"Property '{prop_name}' is mandatory and must not be None"
                    )

        props = self.formatter.format(self.p._all_props())

        if self.is_new:
            request = self._new_attrs()
            request["params"][1][0]["properties"] = props
            resp = self.openbis._post_request(self.openbis.as_v3, request)

            if VERBOSE:
                print("Material successfully created.")
            new_material_data = self.openbis.get_tag(resp[0]["permId"], only_data=True)
            self._set_data(new_material_data)
            return self

        else:
            request = self._up_attrs()
            request["params"][1][0]["properties"] = props
            self.openbis._post_request(self.openbis.as_v3, request)
            if VERBOSE:
                print("Material successfully updated.")
            new_material_data = self.openbis.get_tag(self.permId, only_data=True)
            self._set_data(new_material_data)

    def delete(self, reason="no reason"):
        self.openbis.delete_entity(entity=self.entity, id=self.permId, reason=reason)
        if VERBOSE:
            print(f"Material {self.permId} successfully deleted.")



#   Copyright ETH 2018 - 2024 Zürich, Scientific IT Services
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
#   
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from itertools import chain

from pandas import DataFrame

from .attribute import AttrHolder
from .openbis_object import OpenBisObject
from .things import Things
from .utils import (
    VERBOSE,
    extract_code,
    extract_id,
    extract_nested_identifier,
    extract_userId,
    parse_jackson,
)


class Person(OpenBisObject):
    """managing openBIS persons"""

    def __init__(self, openbis_obj, data=None, **kwargs):
        self.__dict__["openbis"] = openbis_obj
        self.__dict__["a"] = AttrHolder(openbis_obj, "person")

        if data is not None:
            self.a(data)
            self.__dict__["data"] = data

        if kwargs is not None:
            for key in kwargs:
                setattr(self, key, kwargs[key])

    def __dir__(self):
        """all the available methods and attributes that should be displayed
        when using the autocompletion feature (TAB) in Jupyter
        """
        return [
            "permId",
            "userId",
            "firstName",
            "lastName",
            "email",
            "registrator",
            "registrationDate",
            "space",
            "get_roles()",
            "assign_role(role, space)",
            "revoke_role(role)",
        ]

    def get_roles(self, **search_args):
        """Get all roles that are assigned to this person.
        Provide additional search arguments to refine your search.

        Usage::
            person.get_roles()
            person.get_roles(space='TEST_SPACE')
        """
        roles = self.openbis.get_role_assignments(person=self, **search_args)
        groups = self.openbis.get_groups(userId=self.userId, **search_args)

        group_roles = chain.from_iterable(map(lambda x: x["roleAssignments"], groups.response["objects"]))
        count = len(roles) + groups.response["totalCount"]
        response_combined = roles.response["objects"] + list(group_roles)

        return Things(
            openbis_obj=self.openbis,
            entity="role_assignment",
            identifier_name="techId",
            start_with=0,
            count=0,
            totalCount=count,
            response=response_combined,
            df_initializer=self._create_role_assigment_data_frame,
        )

    def _create_role_assigment_data_frame(self, attrs, props, response):
        attrs = ["techId", "role", "roleLevel", "user", "group", "space", "project"]
        if len(response) == 0:
            roles = DataFrame(columns=attrs)
        else:
            objects = response
            parse_jackson(objects)
            roles = DataFrame(objects)
            roles["techId"] = roles["id"].map(extract_id)
            roles["user"] = roles["user"].map(extract_userId)
            roles["group"] = roles["authorizationGroup"].map(extract_code)
            spaces_s = roles["space"].map(extract_code)
            spaces_p = roles["project"].map(lambda x: x['space']['code'] if x is not None else '')
            roles["space"] = spaces_s + spaces_p
            roles["project"] = roles["project"].map(extract_nested_identifier)
        return roles[roles.columns.intersection(attrs)]

    def assign_role(self, role, **kwargs):
        try:
            self.openbis.assign_role(role=role, person=self, **kwargs)
            if VERBOSE:
                print(f"Role {role} successfully assigned to person {self.userId}")
        except ValueError as e:
            if "exists" in str(e):
                if VERBOSE:
                    print(f"Role {role} already assigned to person {self.userId}")
            else:
                raise ValueError(str(e))

    def revoke_role(self, role, space=None, project=None, reason="no reason specified"):
        """Revoke a role from this person."""

        techId = None
        if isinstance(role, int):
            techId = role
        else:
            query = {"role": role}
            if space is None:
                query["space"] = ""
            else:
                if isinstance(space, str):
                    query["space"] = space.upper()
                else:
                    query["space"] = space.code.upper()

            if project is None:
                query["project"] = ""
            else:
                if isinstance(project, str):
                    query["project"] = project.upper()
                else:
                    query["project"] = project.code.upper()

            # build a query string for dataframe
            querystr = " & ".join(f'{key} == "{value}"' for key, value in query.items())
            roles = self.get_roles().df
            if len(roles) == 0:
                if VERBOSE:
                    print(
                        f"Role {role} has already been revoked from person {self.code}"
                    )
                return
            techId = roles.query(querystr)["techId"].values[0]

        # finally delete the role assignment
        ra = self.openbis.get_role_assignment(techId)
        ra.delete(reason)
        if VERBOSE:
            print(f"Role {role} successfully revoked from person {self.code}")
        return

    def __str__(self):
        return f'{self.get("firstName")} {self.get("lastName")}'

    def delete(self, reason):
        raise ValueError("Persons cannot be deleted")

    def save(self):
        if self.is_new:
            request = self._new_attrs()
            resp = self.openbis._post_request(self.openbis.as_v3, request)
            if VERBOSE:
                print("Person successfully created.")
            new_person_data = self.openbis.get_person(resp[0]["permId"], only_data=True)
            self._set_data(new_person_data)
            return self

        else:
            request = self._up_attrs()
            self.openbis._post_request(self.openbis.as_v3, request)
            if VERBOSE:
                print("Person successfully updated.")
            new_person_data = self.openbis.get_person(self.permId, only_data=True)
            self._set_data(new_person_data)



#   Copyright ETH 2018 - 2024 Zürich, Scientific IT Services
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
#   
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from .openbis_object import OpenBisObject
from .utils import is_identifier, is_permid


class Project(OpenBisObject, entity="project", single_item_method_name="get_project"):
    def _modifiable_attrs(self):
        return

    def __dir__(self):
        """all the available methods and attributes that should be displayed
        when using the autocompletion feature (TAB) in Jupyter
        """

        return [
            "add_attachment()",
            "get_attachments()",
            "download_attachments()",
            "get_experiments()",
            "get_samples()",
            "get_datasets()",
            "save()",
            "delete()",
        ] + super().__dir__()

    @property
    def props(self):
        return self.__dict__["p"]

    def get_samples(self, **kwargs):
        return self.openbis.get_samples(project=self.permId, **kwargs)

    get_objects = get_samples  # Alias

    def get_sample(self, sample_code):
        if is_identifier(sample_code) or is_permid(sample_code):
            return self.openbis.get_sample(sample_code)
        else:
            # we assume we just got the code
            return self.openbis.get_sample(project=self, code=sample_code)

    get_object = get_sample  # Alias

    def get_experiments(self):
        return self.openbis.get_experiments(project=self.permId)

    get_collections = get_experiments  # Alias

    def get_datasets(self):
        return self.openbis.get_datasets(project=self.permId)
